import math
from abaqus import *
from abaqusConstants import *
from part import *
from material import *
from section import *
from assembly import *
from step import *
from interaction import *
from load import *
from mesh import *
from optimization import *
from job import *
from sketch import *
from visualization import *
from connectorBehavior import *
from regionToolset import *
from odbAccess import *

class InstanceEdge:
    def __init__(self, model,edge_object):
        self.model=model
        self.edge_object=edge_object
    def get_node_coord_lists(self):
        nodes=[self.model.rootAssembly.instances[self.edge_object.instanceName].vertices[vertix].pointOn for vertix in self.edge_object.getVertices()]
        xs=[]
        ys=[]
        zs=[]
        for node, in nodes:
            x,y,z=node
            xs.append(x)
            ys.append(y)
            zs.append(z)
        return xs,ys,zs
    def get_edge_span(self):
        xs,ys,zs = self.get_node_coord_lists()
        return ([min(xs),max(xs)],[min(ys),max(ys)],[min(zs),max(zs)])
    def get_edge_lengths(self):
        x_span,y_span,z_span = self.get_edge_span()
        return (x_span[1]-x_span[0],y_span[1]-y_span[0],z_span[1]-z_span[0])
    def get_edge_centre_coord(self):
        x_span,y_span,z_span = self.get_edge_span()
        return (sum(x_span)/2, sum(y_span)/2, sum(z_span)/2)
    def get_edge_geom_object(self):
        return self.model.rootAssembly.edges.findAt(coordinates=(((self.get_edge_centre_coord())),))
    def get_direction(self):
        edgelengths=self.get_edge_lengths()
        d=math.sqrt((edgelengths[0]**2+edgelengths[1]**2+edgelengths[2]**2))
        return (edgelengths[0]/d,edgelengths[1]/d,edgelengths[2]/d)
    def detect_overlap(self,coord):
        span = self.get_edge_span()
        for dir_value, span in zip(coord,span):
            if math.floor(min(span))<=dir_value<=math.ceil(max(span)):
                return True
        return False

class GeometryObject(object):
    def __init__(self,origin,lengths):
        self.X,self.Y,self.Z            = origin
        self.length_X,self.length_Y,self.length_Z   = lengths

    def get_bounding_box_coords(self):
        # type: (GeometryObject) -> list[tuple[float]]
        """
        Returns a list of tuples containing the coordinates of the corners of the bounding box
        """
        bounding_box=[
            (
                self.X,
                self.Y,
                self.Z
                ),
            (
                self.X + self.length_X,
                self.Y,
                self.Z + self.length_Z
                ),
            (
                self.X + self.length_X,
                self.Y + self.length_Y,
                self.Z + self.length_Z
                ),
            (
                self.X,
                self.Y + self.length_Y,
                self.Z
                )
        ]
        return bounding_box
    
    def get_bounding_box_span(self):
        # type: (GeometryObject) -> list[tuple[float]]
        """
        Returns a list of tuples containing the ranges of the bounding box
        """
        return(
                [                
                    ((self.X),(self.X + self.length_X)),
                    ((self.Y),(self.Y + self.length_Y)),
                    ((self.Z),(self.Z + self.length_Z))
                ]
            )
    def detect_contact(self,node_coord):
        # type: (GeometryObject, tuple[int]) -> bool
        """
        Detects wheter a given coord touches the bounding box
        """
        bounding_box_span = self.get_bounding_box_span()
        for dir_value, span in zip(node_coord,bounding_box_span):
            if not math.floor(min(span))<=dir_value<=math.ceil(max(span)):
                return False
        return True

    def detect_overlap(self,node_coord):
        # type: (GeometryObject, tuple[int]) -> int
        """
        Counts how many degrees of overlap a contacting coordinate has with the entity's bounding box
        """
        bounding_box_span = self.get_bounding_box_span()
        hits=0
        for dir_value, span in zip(node_coord,bounding_box_span):
            if not math.floor(min(span))<=dir_value<=math.ceil(max(span)):
                return 0
            if math.floor(min(span))<dir_value<math.ceil(max(span)):
                hits+=1
        return hits
    
    def is_on_bounding_box(self,coordinate):
        # type: (GeometryObject, tuple[int]) -> int
        """
        Detects wheter a coordinate is on the perimeter of the entity's bounding box
        """
        own_bounding_box_span = self.get_bounding_box_span()
        hits=0
        for point_axis_coord, bounding_box_axis_span in zip(coordinate, own_bounding_box_span):
            if point_axis_coord in bounding_box_axis_span:
                hits+=1
        return hits>1  
     
    def is_within_projection(self,coordinate):
        # type: (GeometryObject, tuple[int]) -> bool
        """
        Detects whether a point can cast a orthogonal shadow onto the entity's boudning box
        """
        own_bounding_box_span = self.get_bounding_box_span()
        for axis_span, axis_coordinate in zip(own_bounding_box_span,coordinate):
            if axis_span[0]<=axis_coordinate<=axis_span[1]:
                return True
        return False
    
class AssemblyInstance(GeometryObject):
    def __init__(self,model,part,partname,instance,instancename,origin,lengths,section,foundation,distributed_mass,mesh_controls):
        super(AssemblyInstance,self).__init__(origin,lengths)
        self.part                       = part
        self.partname                   = partname
        self.instance                   = instance
        self.instancename               = instancename
        self.model                      = model   
        self.section                    = section
        self.foundation                 = foundation
        self.distributed_mass           = distributed_mass
        self.mesh_controls              = mesh_controls


class ShellInstance(AssemblyInstance):
    def __init__(self,model,part,partname,instance,instancename,origin,lengths,section,foundation,distributed_mass,mesh_controls):
        super(ShellInstance,self).__init__(model,part,partname,instance,instancename,origin,lengths,section,foundation,distributed_mass,mesh_controls)


    def get_instance_edges(self):
        # type: (ShellInstance) -> list[InstanceEdge]
        """Returns a list of InstanceEdge objects associated with the shellInstance"""
        edge_list=[]
        for edge in self.model.rootAssembly.instances[self.instancename].edges:
            edge_list.append(
                InstanceEdge(
                    model=self.model,
                    edge_object=edge,
                )
            )
        return edge_list
    def get_instance_bounding_edges(self):
         # type: (ShellInstance) -> list[InstanceEdge]
        """Returns a list of InstanceEdge objects associated with the shellInstance that lie entirely on the parent instance bounding box"""
        edge_list=[]
        for edge in self.model.rootAssembly.instances[self.instancename].edges:
            edge_instance=InstanceEdge(
                    model=self.model,
                    edge_object=edge,
                )
            if self.is_on_bounding_box(edge_instance.get_edge_centre_coord()):
                edge_list.append(edge_instance)
        return edge_list


class WallInstance(ShellInstance):
    def __init__(self,model,part,partname,instance,instancename,origin,direction,width,height,section,foundation,distributed_mass,mesh_controls):
        length_X=direction[0]*width
        length_Y=height
        length_Z=direction[2]*width 
        super(WallInstance,self).__init__(model,part,partname,instance,instancename,origin,(length_X,length_Y,length_Z),section,foundation,distributed_mass,mesh_controls)


class FloorInstance(ShellInstance):
    def __init__(self,model,part,partname,instance,instancename,origin,width_x,width_z,section,foundation,distributed_mass,mesh_controls):
        length_X=width_x
        length_Y=0
        length_Z=width_z
        super(FloorInstance,self).__init__(model,part,partname,instance,instancename,origin,(length_X,length_Y,length_Z),section,foundation,distributed_mass,mesh_controls)


class WireInstance(AssemblyInstance):
    def __init__(self,model,part,partname,instance,instancename,origin,lengths,section,foundation,distributed_mass,mesh_controls):
        super(WireInstance,self).__init__(model,part,partname,instance,instancename,origin,lengths,section,foundation,distributed_mass,mesh_controls)

    def get_instance_bounding_edges(self):
        edge_list=[]
        for edge in self.model.rootAssembly.instances[self.instancename].edges:
            edge_list.append(
                InstanceEdge(
                    model=self.model,
                    edge_object=edge,
                )
            )
        return edge_list


class Opening(GeometryObject):
    def __init__(self,model,part,partname,instance,instancename,origin,normal,opening_dimension):
        normal                  = normal
        a,b                     = opening_dimension
        length_X                = a*normal[1]+a*normal[2]
        length_Y                = b*normal[0]+b*normal[2]
        length_Z                = a*normal[0]+b*normal[1]

        self.model              = model
        self.part               = part
        self.partname           = partname  
        self.instance           = instance  
        self.instancename       = instancename
        self.normal             = normal
        self.opening_dimension  = opening_dimension
        super(Opening,self).__init__(origin,(length_X,length_Y,length_Z))

class SectionData:
    def __init__(self,name,material,total_thickness,X_thnickess,Y_thickness):
        self.name=name
        self.material=material
        self.total_thickness=total_thickness
        self.X_thickness=X_thnickess
        self.Y_thickness=Y_thickness
        
class ConnectorGroup:
    def __init__(self,model,name,axial_stiffness_ratio,sliding_stiffness_ratio,axial_damping_ratio,sliding_damping_ratio):
        self.model                      = model          
        self.name                       = name
        self.axial_stiffness_ratio = axial_stiffness_ratio
        self.sliding_stiffness_ratio   = sliding_stiffness_ratio
        self.axial_damping_ratio        = axial_damping_ratio
        self.sliding_damping_ratio      = sliding_damping_ratio

    def update_connector_properties(self,model_data):
        if self.name == "Floor-Floor" or self.name=="CB":
            self.set_connector_properties(
                self.model,
                self.name,
                (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
                (0.0, 0.0, 0.0)
                )
        else:
            b_k,h_k=self.get_characteristic_size(instance_bib=model_data["Walls"])
            h,h_x,h_y = model_data["Shell sections"][self.name].total_thickness,model_data["Shell sections"][self.name].X_thickness,model_data["Shell sections"][self.name].Y_thickness
            #Compute the stifnesses for a characteristic panel
            K_X = model_data["Materials"][0]["E0,mean"]*h_x*b_k/h_k #EA/L compression vertical
            K_Y = model_data["Materials"][0]["E0,mean"]*h_y*h_k/b_k #EA/L compression horizontal      
            G_X = model_data["Materials"][0]["G0,mean"]*0.75*h*b_k/h_k #Top and botton 
            G_Y = model_data["Materials"][0]["G0,mean"]*0.75*h*h_k/b_k #Left and right

            num_vertical=math.floor(h_k/model_data["Connector density"]["Vertical"])+1
            num_horizontal=math.floor(b_k/model_data["Connector density"]["Horizontal"])+1

            self.set_connector_properties(
                self.model,
                self.name+" | hor",
                (
                    self.sliding_stiffness_ratio*G_X/num_horizontal,
                    self.axial_stiffness_ratio*K_X/num_horizontal,
                    self.sliding_stiffness_ratio*G_X/num_horizontal,0,0,0),
                (self.sliding_damping_ratio,self.axial_damping_ratio,self.sliding_damping_ratio)
                )
            self.set_connector_properties(
                self.model,
                self.name+" | ver", 
                (
                    self.axial_stiffness_ratio*K_Y/num_vertical,
                    self.sliding_stiffness_ratio*G_Y/num_vertical,
                    self.axial_stiffness_ratio*K_Y/num_vertical,
                    0.0,0.0,0.0),
                (self.axial_damping_ratio,self.sliding_damping_ratio,self.axial_damping_ratio)
                )
        
    def get_characteristic_size(self,instance_bib):
        count=0
        running_width_value=0
        running_height_value=0
        for instance in instance_bib.values():
            if instance.section==self.name:
                count+=1
                running_width_value+=instance.length_X+instance.length_Z
                running_height_value+=instance.length_Y
        try:
            return (running_width_value/count , running_height_value/count)
        except:
            return (1,1)
    def set_connector_properties(self,model,name,stiffnesses,damping):
        unlocked_DOF={}
        locked_DOF={}
        for idx, DOFstiffness in enumerate(stiffnesses):
            if DOFstiffness == 0:
                locked_DOF[idx+1]=DOFstiffness
            else:
                unlocked_DOF[idx+1]=DOFstiffness
        behaviour_options=[]
        if len(unlocked_DOF.keys()) >0:
            behaviour_options.append(
                ConnectorElasticity(
                    table=(tuple(unlocked_DOF.values()), ), 
                    independentComponents=(), 
                    components=tuple(unlocked_DOF.keys())
                    ),         
                )
        if len(locked_DOF.keys()) >0:
            behaviour_options.append(
                ConnectorElasticity(
                    behavior=RIGID, 
                    table=((0.0, 0.0), ), 
                    independentComponents=(), 
                    components=tuple(locked_DOF.keys())
                    ),         
                )
        if sum(damping)>0:
            behaviour_options.append(
                ConnectorDamping(
                    table=((damping), ), 
                    independentComponents=(), 
                    components=(1, 2, 3)
                    )
            )
        model.sections[name].setValues(
            behaviorOptions=tuple(behaviour_options)
            )

